import collections
import io
import json
import logging
import pprint
import os
import threading
import time
import tempfile
from typing import Dict, Any

import requests
import string
import uuid
import urllib3
import zipfile
import copy

urllib3.disable_warnings() # HP uses self-signed certs, this produces a warning on every request - disable them

import base_sender
import job_report
import log
import paths
import xmltodict


class Sender(base_sender.BaseSender):

    EXTENDED_DATA = True
    REPORT_JOBS = True
    REPORT_MATERIALS = True
    REPORT_BUILD_UNIT = True

    CONNECTION_TIMEOUT = 30 #don't reduce this, HP is too slow - you can provoke connection errors
    HTTP_TIMEOUT = 10
    UPDATE_LOOP_SLEEP = 6
    UPDATE_LOOP_STEPS = 100
    MONITORING_LOOP_SLEEP = 3
    UPLOAD_PROCESSING_TIMEOUT = 5*60 # 5 minutes 
    TWENTY_THREE_HOURS_IN_SECONDS = 60*60*23
    HOST_MASK = "http://%s:%d"
    HTTPS_HOST_MASK = "https://%s:%d"
    VERIFY_SSL = False
    REQUEST_RETRIES = 3
    REQUEST_RETRY_WAIT = 5

    HP_API_NS_NAME = "hpmjf1" 
    HP_API_NS_VALUES = ["http://www.hp.com/schemas/hpmjf/v1_" + str(i) for i in (2,3,4,5)]

    APP_NAME = "3DPrinterOS"

    JOB_PROFILES_NAMES = [
        "PrintProfiles",
        "CoolingProfiles",
        "AnnealingProfiles",
        "WarmingProfiles",
        "CuringProfiles"
    ]

    #PRINT_STATUSES = ("JobChecking", "JobPreparingToPrint", "JobWarmingUp", "JobPrinting", "JobAnnealing", "JobCooling", "JobCancelling")
    #use our statuses instead of HP's 
    PRINT_STATUSES = ("checking", "waiting_to_print", "preparing", "warming", "printing", "annealing")
    PRINT_ACTIVITIES = ("ActivityPrinting", "ActivityWarmingUp", "ActivityAnnealing")
    CREATE_JOB_TEMPLATE_FILE = os.path.join(paths.APP_FOLDER, "hp_create_job_template.xml")
    SECRET_FILENAME = "hsep.bin"
    # don't user os.path.join for this path, zipfile paths will always as in unixlike format with folder as "/"
    MATERIALS_AND_PROFILES_FILE_IN_3MF = "Metadata/3dprinteros_hp_materials.json" 

    LOG_ALL = False
    #NOTE comment this is production
    #DEBUG_JOB_XML_PATH = "/tmp/my_job_ticket.xml"
    SUPPORT_JOBS = True

    def __init__(self, parent, usb_info, profile={}):
        super().__init__(parent, usb_info, profile)
        self.stop_flag = False
        self.logger = logging.getLogger(__name__)
        self.profile = profile # playing no role in this sender
        self.run_once = usb_info.get('RUNONCE', False) # present when used by http_detector
        host = usb_info.get("IP")
        port = usb_info.get("PORT")
        if not host or not port:
            raise RuntimeError("Can't connection to HP Gateway with %s" % self.usb_info)
        self.status = "connecting"
        self.operational_flag = False
        self.printing_flag = False
        self.cooling_flag = False
        self.paused_flag = False
        self.percent = 0
        self.total_lines = 0
        self.current_line_number = 0
        self.discovery_host = self.HOST_MASK % (host, port)
        self.discovery_host_https = self.HTTPS_HOST_MASK % (host, 8443)
        self.deviceproxy_host = ""
        self.auth_key = ""
        self.device_id = ""
        self.identification_dict = {}
        #self.jobs = []
        self.jobs_by_id = {}
        self.job_statuses = {}
        self.printers_job_id = ""
        self.clouds_job_id = ""
        self.clouds_job_snr = ""
        self.autostart_job = False
        self.materials = {}
        self.default_material_id = ""
        self.selected_material_id = ""
        self.selected_profiles_per_material = {}
        self.serial_number = usb_info.get("SNR", "")
        self.model_name = ""
        self.model_number = ""
        self.part_number = ""
        self.manufacturer = ""
        self.friendly_name = ""
        self.device_region = ""
        self.fw_release_name = ""
        self.fw_release_date = ""
        self.user_name = self.parent.parent.user_login.login
        self.local_ip = self.parent.parent.lan_ip
        self.build_units = {}
        self.can_load_build_unit = False
        self.loaded_build_unit = None
        self.maintenance_flag = False
        if self.LOG_ALL:
            self.logger.warning("Warning! Logging all responses dicts to /tmp/. Use this only in debug!")
        self.connect()

    @staticmethod
    def remove_hp_namespace_junk(input_string):
        output = input_string.replace(Sender.HP_API_NS_NAME + ":", "") 
        for replacement_line in Sender.HP_API_NS_VALUES:
            output = output.replace(replacement_line + ":", "")
        return output
        #return input_string.replace(Sender.HP_API_NS_NAME + ":", "").replace(Sender.HP_API_NS_VALUE + ":", "") 

    def connect(self):
        if not self.run_once:
            self.connect_thread = threading.Thread(target=self.connect_and_status_loop, name="HP_" + self.serial_number)
            self.connect_thread.start()
            self.status_thread_crash_monitoring_thread = threading.Thread(target=self.status_crash_monitoring, name="HP_crash_mon_" + self.serial_number)
            self.status_thread_crash_monitoring_thread.start()
            steps = self.UPDATE_LOOP_STEPS
            while steps:
                time.sleep(self.CONNECTION_TIMEOUT/self.UPDATE_LOOP_STEPS)
                steps -= 1
                if self.operational_flag:
                    return
            self.stop_flag = True
            raise RuntimeError("No connection to %s" % self.usb_info)

    def status_crash_monitoring(self):
        while not self.stop_flag:
            if self.connect_thread:
                if not self.connect_thread.is_alive():
                    self.operational_flag = False
            time.sleep(self.MONITORING_LOOP_SLEEP)

    def select_material_and_profiles(self, materials_and_profiles_dict):
        material_dict = materials_and_profiles_dict.get('material')
        if isinstance(material_dict, dict):
            material_id = material_dict.get('id')
            material_name = material_dict.get('name')
            profiles = material_dict.get('profiles')
            self.logger.info("Selecting material %s %s" % (material_name, material_id))
            self.selected_material_id = material_id
            self.selected_profiles_per_material[material_id] = profiles
            return True
        return False

    def parse_material_and_profiles_from_3mf(self, file_3mf):
        try:
            with zipfile.ZipFile(file_3mf) as zf:
                with zf.open(self.MATERIALS_AND_PROFILES_FILE_IN_3MF, 'r') as f:
                    json_text = f.read()
        except (OSError, IOError):
            self.parent.register_error(5117, "File error on opening 3mf to select materials", is_blocking=False)
        except MemoryError:
            self.parent.register_error(5118, "Out of memory while opening 3mf", is_blocking=False)
        except zipfile.error:
            self.parent.register_error(5119, "Error: 3mf is not a zipfile", is_blocking=False)
        except KeyError:
            self.logger.warning("No materials file in the 3mf")
        else:
            try:
                return self.select_material_and_profiles(json.loads(json_text))
            except Exception as e:
                self.logger.warning("Exception on parsing material json from 3mf: %s\nText: %s" % (str(e), str(json_text)))
        return False

    def gcodes(self, gcodes_or_file):
        if isinstance(gcodes_or_file, io.IOBase):
            return self.upload_to_printer(gcodes_or_file, self.autostart_job)
        else:
            self.unbuffered_gcodes(gcodes_or_file)
        return False

    def upload_to_printer(self, file_to_upload, autostart=False):
        if not self.parse_material_and_profiles_from_3mf(file_to_upload):
            self.logger.warning("Using default materials and profiles.")
        upload_url = self.create_job()
        if not upload_url:
            self.logger.warning(f"Job creation failed.")
            return False
        else:
            return self.upload_job_file(file_to_upload, upload_url, autostart)
           
    def unbuffered_gcodes(self, gcodes):
        self.parent.register_error(505, "Console and joystick are not supported for this printer type", is_blocking=False)
        return False

    def is_printing(self):
        return self.printing_flag

    def is_paused(self):
        return self.paused_flag

    def is_operational(self):
        return self.operational_flag

    def is_bed_not_clear(self):
        return self.loaded_build_unit is not None and self.cooling_flag

    def is_empty(self):
        return self.can_load_build_unit and self.loaded_build_unit is None

    def is_maintenance(self):
        return self.maintenance_flag

    def pause(self):
        self.parent.register_error(505, "Pause is not supported", is_blocking=False)
        return False

    def unpause(self):
        self.parent.register_error(505, "Pause is not supported", is_blocking=False)
        return False

    def cancel(self):
        if self.is_printing():
            if self.printers_job_id:
                return self.cancel_job(self.printers_job_id)
        return False

    def get_percent(self):
        if self.is_printing():
            return self.percent
        return 0

    def get_current_line_number(self):
        if self.is_printing():
            return self.current_line_number
        return 0

    def get_total_gcodes(self):
        if self.is_printing():
            return self.total_lines
        return 0

    def get_jobs(self):
        jobs = list(self.jobs_by_id.values())
        self.logger.debug("Reported jobs:\n" + pprint.pformat(jobs))
        return jobs

    def get_materials(self):
        materials_list = []
        try:
            materials_keys_list = list(self.materials.keys())
            materials_keys_list.sort()
            for key in materials_keys_list:
                material = copy.deepcopy(self.materials[key])
                if material.get('xml'):
                    del material['xml']
                for profile_type_name in material.get('profiles', {}):
                    for profile_dict in material['profiles'][profile_type_name]:
                        if profile_dict.get('xml'):
                            del profile_dict['xml']
                materials_list.append(material)
            self.logger.debug("Reported materials:\n" + pprint.pformat(materials_list))
        except Exception as e:
            self.logger.exception("Exception in creation of materials report: %s" % e)
        return materials_list

    def get_build_unit(self):
        if self.loaded_build_unit is not None and self.build_units[self.loaded_build_unit] is not None:
            return self.build_units[self.loaded_build_unit]
        return {}

    def load_username_and_password(self):
        for secret_file_path in (os.path.join(paths.APP_FOLDER, self.SECRET_FILENAME),\
                os.path.join(paths.CURRENT_SETTINGS_FOLDER, self.SECRET_FILENAME)):
            if os.path.isfile(secret_file_path):
                try:
                    with open(secret_file_path, "rb") as f:
                        secret = f.read()
                except Exception as e:
                    self.logger.warning("No HP credentials found in :%s ", str(secret_file_path))
                else:
                    secret = json.loads(bytes([x ^ y for x, y in zip(secret, b'glasswoodentin'*10)]))
                    if isinstance(secret, dict):
                        return secret.get('username'), secret.get('password')
        self.logger.warning("Unable to load HP credentials!")
        return None, None

    def req(self, url, post=False, add_auth_header=True, **kwargs):
        retries_left = self.REQUEST_RETRIES 
        if post:
            method = 'POST'
        else:
            method = 'GET'
        if add_auth_header:
            kwargs["headers"] = {"Authorization" : self.auth_key}
        while retries_left:
            try:
                r = requests.request(method, url, **kwargs, verify=self.VERIFY_SSL, timeout=self.HTTP_TIMEOUT)
            except requests.RequestException as e:
                self.logger.debug("Error requesting url %s. %s" % (url, e))
                r = None
                retries_left -= 1
                time.sleep(self.REQUEST_RETRY_WAIT)
            else:
                if not r.ok:
                    self.logger.warning("Warning invalid response on %s: %s, %s" % (url, r.status_code, r.content))
                return r

    def req_get(self, url, **kwargs):
        return self.req(url, post=False, **kwargs)

    def req_post(self, url, **kwargs):
        return self.req(url, post=True, **kwargs)

    def req_xml_as_dict(self, url, post=False, **kwargs):
        if not self.auth_key:
            self.logger.info("No auth header to perform request")
        else:
            auth_header = self.get_auth_headers()
            retries_left = self.REQUEST_RETRIES 
            while retries_left:
                try:
                    if post:
                        r = requests.post(url, headers=auth_header, **kwargs, verify=self.VERIFY_SSL, timeout=self.HTTP_TIMEOUT)
                    else:
                        r = requests.get(url, headers=auth_header, verify=self.VERIFY_SSL, timeout=self.HTTP_TIMEOUT)
                except Exception as e:
                    self.logger.warning("Error when requesting %s. Error: %s" % (url, e))
                    retries_left -= 1
                    time.sleep(self.REQUEST_RETRY_WAIT)
                else:
                    if r.ok:
                        try:
                            text = r.text
                            if not isinstance(text, str):
                                raise TypeError
                            text = self.remove_hp_namespace_junk(text)
                            resp_dict = xmltodict.parse(text)
                        except:
                            self.logger.error("Error parsing XML. URL: %s" % url)
                        else:
                            if self.LOG_ALL:
                                #formatted_dict = pprint.pformat(resp_dict)
                                #self.logger.info("LOG_ALL(!Task):" + formatted_dict)
                                #self.logger.info("LOG_ALL: " + formatted_dict)
                                self.logger.info("LOG_ALL:\n%s\n=====" % text)
                            return resp_dict
                        finally:
                            r.close()
                    else:
                        self.logger.warning("Error when requesting URL: %s\nStatus: %s\nResponse: %s\n" % (url, r.status_code, r.content))
                        r.close()
                        return

    @log.log_exception
    def connect_and_status_loop(self):
        self.request_device_proxy(self.discovery_host_https + "/discovery")
        if not self.deviceproxy_host:
            self.request_device_proxy(self.discovery_host + "/discovery")
            if not self.deviceproxy_host:
                self.logger.warning('Cant get DeviceProxy from Discovery')
                return 
        self.authenticate()
        if not self.auth_key:
            self.logger.error('Error on authentication')
            return 
        self.request_device_id()
        self.request_identification()
        self.request_materials()
        self.request_bed_constraints()
        auth_time = time.monotonic()
        while not self.stop_flag:
            self.request_status()
            self.request_trolleys()
            self.request_jobs()
            #self.request_sensors()
            steps = self.UPDATE_LOOP_STEPS
            while steps:
                time.sleep(self.UPDATE_LOOP_SLEEP/self.UPDATE_LOOP_STEPS)
                steps -= 1
            if time.monotonic() - auth_time > self.TWENTY_THREE_HOURS_IN_SECONDS:
                self.refresh_auth()
                auth_time = time.monotonic()

    def request_device_proxy(self, url):
        self.logger.debug("Requesting discovery on URL: %s" % url)
        r = self.req_get(url, add_auth_header=False)
        if r and r.ok:
            try:
                response_dict = json.loads(r.text)
            except:
                self.logger.warning("Discovery response is invalid json: %s" % r.text)
            else:
                self.logger.debug("Discovery response: " + pprint.pformat(response_dict))
                proxy = response_dict.get('links', {}).get("DeviceProxy")
                apigate = response_dict.get('links', {}).get("ApiGateway")
                if proxy and isinstance(proxy, dict):
                   proxy_url = proxy.get("href")
                   if isinstance(proxy_url, str):
                       self.deviceproxy_host = proxy_url
                if apigate and isinstance(apigate, dict):
                   apigate_url = apigate.get("href")
                   if isinstance(apigate_url, str):
                       self.apigate_host = apigate_url
            finally:
                r.close()

    def authenticate(self):
        if not self.deviceproxy_host:
            self.logger.warning("No DeviceProxy URL to perfom authentication")
            return
        username, password = self.load_username_and_password()
        auth_dict = {"username": username, "password": password}
        url = self.deviceproxy_host + "/authenticate"
        self.logger.debug('Requesting authentication on URL:%s' % url)
        r = self.req_post(url, json = auth_dict, add_auth_header = False)
        if r:
            if r.status_code == 400:
                self.logger.warning('Invalid HPGatewayAPI credentials')
            elif r.ok:
                auth_key = r.text
                try:
                    r.close()
                except:
                    pass
                if isinstance(auth_key, str):
                    self.logger.debug("Successful authentication on HP GatewayAPI Server")
                    self.auth_key = auth_key
                    self.logger.warning(f"Auth key: {self.auth_key}")

    def refresh_auth(self):
        if self.deviceproxy_host:
            r = self.req_get(self.deviceproxy_host + "/authentication/refresh")
            if r:
                r.close()

    def get_auth_headers(self):
        return {"Authorization" : self.auth_key}

    def get_devices(self, deviceproxy_host=None):
        if deviceproxy_host:
            url = deviceproxy_host + "/devices"
        else:
            url = self.deviceproxy_host + "/devices"
        devices_dict = self.req_xml_as_dict(url)
        returned_devices = {}
        if isinstance(devices_dict, dict):
            services_dict = devices_dict.get("ServiceDiscovery")
            if isinstance(services_dict, dict):
                devices = services_dict.get("Devices")
                if isinstance(devices, dict):
                    for device_name in devices:
                        devices_list = devices[device_name]
                        if not isinstance(devices_list, list):
                            devices_list = [devices_list]
                        for device in devices_list:
                            device_type = device.get("@type")
                            if device_type == "printer":
                                device_id = device.get("@id")
                                if device_id:
                                    device_name = device.get("@name")
                                    returned_devices[device_id] = {"name" : device_name}
        for device_id in returned_devices:
            iden_dict = self.get_identification(device_id)
            if iden_dict:
                self.logger.info("Printer identification received for %s:\n %s" % (device_id, pprint.pformat(iden_dict)))
                returned_devices[device_id].update(iden_dict)
        return returned_devices

    def request_device_id(self, wanted_snr=None):
        devices = self.get_devices()
        for device_id in devices:
            identification = self.get_identification(device_id)
            if identification:
                snr = identification.get('serial_number')
                if not wanted_snr:
                    wanted_snr = self.serial_number
                    if not wanted_snr:
                        self.logger.info("Printer ID:[%s]\tPrinter SNR:[%s]" % (device_id, snr))
                        continue
                if wanted_snr == snr:
                    self.device_id = device_id
                    self.logger.debug("Got device id: %s" % self.device_id)
                    return
        if not self.run_once:
            self.logger.warning("Error getting device id by serial number %s. Devices identification request didn't returned a device with such id" % wanted_snr)

    def service_discovery(self, device_id, method="Get"):
        services = {}
        url_mask = self.deviceproxy_host + "/devices/%s/servicediscovery"
        services_dict = self.req_xml_as_dict(url_mask % device_id)
        if services_dict:
            services_dict = services_dict.get("ServiceDiscovery", {})
            services_list = services_dict.get("SupportedIfc") 
            for service in services_list:
                service_name = service.get("@resource")
                if service_name:
                    urls = service.get("ResourceURI", {})
                    if not isinstance(urls, list):
                        urls = [urls]
                    for url_dict in urls:
                        dicts_method = url_dict.get("@method", "")
                        if dicts_method and dicts_method != method:
                            continue
                        url = url_dict.get("@uri", "")
                        if url:
                            #self.logger.debug("Got service %s to printer %s" % (service_name, device_id))
                            services[service_name] = url
        return services

    def request_identification(self):
        id_dict = self.get_identification(self.device_id)
        if id_dict:
            self.model_name = id_dict.get('model_name')
            self.model_number = id_dict.get('model_number')
            self.part_number = id_dict.get('part_number')
            self.manufacturer = id_dict.get('manufacturer')
            self.serial_number = id_dict.get('serial_number')
            self.friendly_name = id_dict.get('friendly_name')
            self.device_region = id_dict.get('device_region')
            self.fw_release_name = id_dict.get('fw_release_name')
            self.fw_release_date = id_dict.get('fw_release_date')
            self.logger.debug("Printer identification data updated")
        else:
            self.logger.warning("Error getting printer identification")

    def request_bed_constraints(self):
        if not self.identification_dict:
            self.logger.warning("No identification to parse bed min and max from")
        self.parse_default_bed_from_identification(self.identification_dict)
        self.parse_materials_from_identification(self.identification_dict)

    def get_identification(self, device_id=None):
        if device_id:
            if self.identification_dict and self.device_id == device_id:
                identification_dict = self.identification_dict
            else:
                identification_dict = self.get_service_as_dict('identification', device_id=device_id)
                if identification_dict:
                    identification_dict = identification_dict.get('Identification')
                    #self.logger.debug("Identification: %s" + pprint.pformat(identification_dict))
            #self.logger.debug("ID dict:" + pprint.pformat(identification_dict))
            if identification_dict:
                if device_id == self.device_id:
                    self.identification_dict = identification_dict
                return self.parse_identification_fields(identification_dict)
        self.logger.warning("No device id to get identification")

    def parse_identification_fields(self, identification_dict):
        id_dict = {}
        fields = identification_dict.get('Fields')
        if isinstance(fields, dict):
            id_dict["model_name"] = fields.get('ModelName')
            id_dict["model_number"] = fields.get('ModelNumber')
            id_dict["part_number"] = fields.get('PartNumber')
            id_dict["manufacturer"] = fields.get('Manufacturer')
            id_dict["serial_number"] = fields.get('SerialNumber')
            id_dict["friendly_name"] = fields.get('FriendlyName')
            id_dict["device_region"] = fields.get('DeviceRegion')
            id_dict["fw_release_name"] = fields.get('FwReleaseName')
            id_dict["fw_release_date"] = fields.get('FwReleaseDate')
        return id_dict

    def parse_default_bed_from_identification(self, identification_dict):
        builds_platform_dict = identification_dict.get("PrinterProperties", {}).get("BuildPlatforms", {}).get("BuildPlatform", {})
        if not isinstance(builds_platform_dict, dict):
            self.logger.warning("Empty BuildPlatform field in identification")
            return
        min_max_dict = self.parse_coords_min_max_from_p1p2(builds_platform_dict)
        if min_max_dict:
            min_max_dict['id'] = "default"
            min_max_dict['name'] = "default"
            self.default_bed_min_max_coords = min_max_dict
        self.logger.info("Default bed constraints:\n" + pprint.pformat(min_max_dict))

    def parse_materials_from_identification(self, identification_dict):
        materials_list = identification_dict.get("Materials", {}).get("Material")
        if not materials_list:
            self.logger.warning("Empty materials list")
        else:
            if not isinstance(materials_list, list):
                materials_list = [materials_list]
            for material_dict in materials_list:
                if not isinstance(material_dict, dict):
                    self.logger.warning("Non dictionary entity parsed from identification/materials: %s" % material_dict)
                    continue
                material_name = material_dict.get("Name", {}).get("#text", "")
                material_id = material_dict.get("ID", "")
                min_max_coords = self.parse_coords_min_max_from_p1p2(material_dict.get("BuildPlatform"))
                if material_dict.get("Default", "").lower() == "true":
                    min_max_coords["default"] = True
                    self.default_material_id = material_id
                #store into subdict of self.materials
                self.materials.get(material_id, {}).update(min_max_coords)
                self.logger.info(f"Material {material_name} {material_id} bed constraints:\n" + pprint.pformat(min_max_coords))
        #self.logger.debug("Available materials: \n" + pprint.pformat(self.materials))

    def parse_coords_min_max_from_p1p2(self, build_platform_dict):
        usable_platform_dict = build_platform_dict.get('UsablePlatform')
        if not isinstance(usable_platform_dict, dict):
            self.logger.warning("Empty usable platform dict")
            return {}
        else:
            self.logger.debug(f"Usable platform in HP format: {usable_platform_dict}")
            p1_dict = usable_platform_dict.get("P1", {})
            p2_dict = usable_platform_dict.get("P2", {})
            units = usable_platform_dict.get("Units")
            if units == "micron":
                multiplier = 0.001
            elif units == "mm":
                multiplier = 1
            elif units == "cm":
                multiplier = 10
            else:
                self.logger.warning("Unknown or absent Units field in UsablePlatform: %s. Assuming its microns." % units)
                multiplier = 0.1
            try:
                min_max_dict = {}
                for prefix, coords_dict in (("min_", p1_dict), ("max_", p2_dict)):
                    for coord_name in ("X", "Y", "Z"):
                        min_max_dict[prefix + coord_name.lower()] = float(coords_dict[coord_name]) * multiplier
            except Exception as e:
                #self.logger.warning("")
                self.logger.exception("Exception %s while parsing min and max coords for usable platform" % e)
            return min_max_dict

    def request_materials(self):
        materials_dict = self.get_service_as_dict('materials')
        if isinstance(materials_dict, dict):
            materials_list = materials_dict.get('Materials', {}).get('Material', {})
            if not materials_list:
                self.logger.warning("Warning. No materials")
                return
            if not isinstance(materials_list, list):
                materials_list = [materials_list,]
            for material in materials_list:
                material_id = material.get('ID')
                if isinstance(material_id, str):
                    material_name = material.get('Name', {}).get("#text")
                    if isinstance(material_name, str):
                        material_details_url = material.get('Links', {}).get("Link", {}).get("@uri", "")
                        if material_details_url:
                            resp = self.req_get(material_details_url)
                            if resp and resp.ok:
                                xml_text = resp.text
                            else:
                                xml_text = ""
                            if isinstance(xml_text, str):
                                our_material_dict = {}
                                our_material_dict['id'] = material_id
                                our_material_dict['name'] = material_name
                                our_material_dict['xml'] = xml_text 
                                profile_xmls_by_type = self.get_profiles_from_material_xml(xml_text)
                                if profile_xmls_by_type and isinstance(profile_xmls_by_type, dict):
                                    our_material_dict['profiles'] = self.get_profiles_data_from_profiles_xml(profile_xmls_by_type, material_name)
                                self.materials[material_id] = our_material_dict
                material_debug_text = pprint.pformat(material) + "\n"
                try:
                    material_debug_text = material_debug_text + pprint.pformat(xmltodict.parse(xml_text))
                except:
                    material_debug_text = material_debug_text + "Error parsing XML"
                #self.logger.debug("Material %s: %s" % (material_name, material_debug_text))
            self.logger.info("Updated of printer's materials. Got %s." % len(self.materials))

    def get_profiles_data_from_profiles_xml(self, profiles_xml_dict, material_name=""):
        our_format_profiles_per_type_dict = {}
        for profile_type_name in profiles_xml_dict:
            profiles_xmls_list = profiles_xml_dict.get(profile_type_name)
            if profiles_xmls_list:
                our_format_profiles_per_type_dict[profile_type_name] = []
                if not isinstance(profiles_xmls_list, list):
                    profiles_xmls_list = [profiles_xmls_list]
                for profile_xml_with_namespaces in profiles_xmls_list:
                    profile_xml = self.remove_hp_namespace_junk(profile_xml_with_namespaces)
                    profile_dict = xmltodict.parse(profile_xml)
                    if profile_dict and isinstance(profile_dict, dict):
                        profile_dict = profile_dict.get(profile_type_name[:-1])
                        if profile_dict:
                            our_format_profile_dict = { 'xml': profile_xml_with_namespaces,\
                                            'name': profile_dict.get('Name', {}).get("#text", ""),\
                                            'id': profile_dict.get('ID', "") }
                            our_format_profiles_per_type_dict[profile_type_name].append(our_format_profile_dict)
                            #self.logger.debug("Appending %s to material %s " % (pprint.pformat(our_format_profile_dict), material_name))

                        else:
                            self.logger.warning("Profile format is broken: %s" % pprint.pformat(profile_dict))
                    else:
                        self.logger.warning("Invalid result of profile parsing - None or not dict: %s" % profile_dict)
        return our_format_profiles_per_type_dict

    #  def service_discovery_all_printers(self):
    #      for device_id in self.devices:
    #          for method in ("Get", "Post"):
    #              services = self.service_discovery(device_id, method=method)
    #              if services:
    #                  services['method'] = method
    #                  self.devices[device_id]["services"].update(services)

    def get_service_as_dict(self, service_name, device_id=None):
        if not device_id:
            device_id = self.device_id
        if not device_id:
            self.logger.warning("No device_id to get services")
        else:
            services = self.service_discovery(device_id, method="Get")
            if isinstance(services, dict):
                service_uri = services.get(service_name)
                if service_uri:
                    service_dict = self.req_xml_as_dict(service_uri)
                    if isinstance(service_dict, dict):
                        return service_dict
                else:
                    self.logger.error('Error getting service url for %s. No such service.' % service_name)
                    return {}

    def request_status(self):
        devicestatus = self.get_service_as_dict("devicestatus")
        if devicestatus:
            devicestatus = devicestatus.get("DeviceStatus")
            if isinstance(devicestatus, dict):
                self.logger.debug("Status: %s" % pprint.pformat(devicestatus))
                printing = False
                activities = devicestatus.get("OngoingActivities")
                if isinstance(activities, dict):
                    for activity_name in activities:
                        try:
                            activity_name = activities.get(activity_name, {}).get("Name", {}).get("#text", "")
                        except AttributeError:
                            self.logger.warning("Error parsing printer status by activities")
                            continue
                        if activity_name in self.PRINT_ACTIVITIES:
                            printing = True
                        self.maintenance_flag = activity_name == "ActivityDeviceMaintenance"
                        self.cooling_flag = activity_name == "ActivityCooling"
                        self.logger.info("Device ongoing activity: %s" % activity_name)
                self.printing_flag = printing
                summary = devicestatus.get("Summary", {}).get("#text")
                if summary == "DeviceStatusWithAlerts":
                    self.get_alerts()
                operational_flag = True
                severity = devicestatus.get("Severity", {}).get("#text")
                if severity == "DeviceStatusSeverityError":
                    self.logger.warning("Received severe error status")
                    operational_flag = False
                self.operational_flag = operational_flag
                if self.operational_flag:
                    self.logger.debug("Status operational")
                return 
        self.operational_flag = False
        self.logger.warning("Error updating device status. Device status:%s" % devicestatus)

    def request_trolleys(self):
        trolleys = self.get_service_as_dict("trolleys")
        if trolleys:
            self.can_load_build_unit = True
            self.loaded_build_unit = None
            trolleys_dict = trolleys.get("Trolleys", {})
            trolley = trolleys_dict.get("Trolley", [])
            if isinstance(trolley, dict):
                self.handle_trolley_data(trolley)
            elif isinstance(trolley, list):
                for trolley_dict in trolley:
                    self.handle_trolley_data(trolley_dict)

    def handle_trolley_data(self, trolley_dict):
        if trolley_dict.get("IsPresent") != "true":
            return
        description = trolley_dict.get("Description", {})
        if description.get("Type", {}).get("#text") != "TrolleyTypeBuildUnit":
            return
        uuid = description.get("UUID")
        if self.build_units.get(uuid) is None:
            bu_dict = {}
            bu_dict["id"] = trolley_dict.get("ID")
            bu_dict["name"] = description.get("Name")
            bu_dict["sn"] = description.get("SerialNumber")
            self.build_units[uuid] = bu_dict
        status_dict = trolley_dict.get("Status", {})
        if status_dict:
            self.loaded_build_unit = uuid
            status_new = {}
            status_new["status"] = status_dict.get("StatusMessage", {}).get("#text", "")
            status_new["load_percent"] = status_dict.get("LoadPercentage")
            status_new["recycled_percent"] = status_dict.get("RecycledPercentage")
            self.build_units[uuid].update(status_new)
            material_dict = status_dict.get("Material")
            if material_dict is None:
                return
            if self.build_units[uuid].get("material", {}).get("id") != material_dict.get("ID"):
                material_new = {"id": material_dict.get("ID"), "name": material_dict.get("Name", {}).get("#text")}
                self.build_units[uuid].update({"material": material_new})

    def get_alerts(self):
        alerts = self.get_service_as_dict("alerts")
        if alerts:
            alerts = alerts.get("Alerts", {})
            for alert in alerts.get("Alert", []):
                if isinstance(alert, dict):
                    message = alert.get("Message", {}).get("#text")
                    is_blocking = False
                    is_info = False
                    severity = alert.get("Severity", "")
                    if severity == "Error":
                        #is_blocking = True
                        #self.operational_flag = False #TODO check how it will be of use
                        self.logger.warning("Printer has error alert: %s" % message)
                    elif severity == "Warning":
                        is_blocking = False
                    elif severity == "Informative":
                        is_info = True
                    else:
                        self.logger.warning("Unknown severity: %s" % severity)
                    self.parent.register_error(2021, message, is_blocking=is_blocking, is_info=is_info)
            return
        self.logger.warning("Error updating alerts. Alerts: %s" % alerts)

    def get_job_name(self):
        postfix = ""
        if self.clouds_job_snr:
            postfix += "SN_" + self.clouds_job_snr + "_"
        if self.clouds_job_id:
            postfix += "Job_" + self.clouds_job_id
        else:
            postfix = str(uuid.uuid1())
        return "ZAP_" + postfix

    def request_jobs(self):
        jobs = self.get_service_as_dict("jobs")
        if not jobs:
            self.logger.warning("Error updating jobs. Jobs: %s" % jobs)
            return
        jobs_list = jobs.get("Jobs", {}).get("Job", [])
        if not isinstance(jobs_list, list):
            jobs_list = [jobs_list]
        active_job_id = ""
        current_job_ids = []
        for job_dict in jobs_list:
            #pprint.pprint(("!Job dict", job_dict))
            job_id = job_dict.get("Job_ID")
            if not job_id:
                self.logger.warning("Warning at processing job: %s" % job_dict)
                continue
            current_job_ids.append(job_id)
            job_status = job_dict.get("Status", {}).get("Message", {}).get("#text", "")
            is_printing = job_report.hp_job_status_to_job_status(job_status) in self.PRINT_STATUSES
            if not is_printing and self.job_statuses.get(job_id) is not None and self.job_statuses.get(job_id) == job_status:
                #Skip job if not printing and status was not changed
                continue
            new_job_dict = {"Status": job_status, "JobID": job_id}
            details_uri = job_dict.get("Links", {}).get("Link", {}).get("@uri", "")
            details_dict = self.req_xml_as_dict(details_uri)
            if details_dict:
                new_job_dict.update(details_dict.get("Job", {}))
            status_uri = details_uri + "/status"
            status_dict = self.req_xml_as_dict(status_uri)
            if status_dict:
                progress_dict = status_dict.get("JobStatus", {}).get("UserReported", {}).get("Progress", {})
                new_job_dict.update(progress_dict)
            if new_job_dict.get("Trolley_ID") is not None:
                if self.build_units.get(new_job_dict.get("Trolley_ID")) is not None:
                    new_job_dict.update({"trolley_sn": self.build_units.get(new_job_dict.get("Trolley_ID")).get("sn")})

            parsed_job_dict = dict(job_report.HPJobReport(new_job_dict))
            if parsed_job_dict:
                if is_printing:
                    if active_job_id:
                        self.logger.warning("Warning! Printer reported more than one job with active states")
                    else:
                        active_job_id = parsed_job_dict.get("id", "")

                        total_lines = parsed_job_dict.get('layers', 0)
                        current_line_number = parsed_job_dict.get('current_layer', 0)
                        status = parsed_job_dict.get("status", "")
                        percent = parsed_job_dict.get('progress', 0) # Simulator return wrong progress percent
                        # WORKAROUND: Calculate percent from layers info, HP simulators returns wrong progress percent.
                        #if self.total_lines > 0:
                        #    self.percent = int(self.current_line_number / self.total_lines * 100)
                self.jobs_by_id[job_id] = parsed_job_dict
                self.job_statuses[job_id] = job_status
            else:
                self.logger.debug("Error parsing job: %s", pprint.pformat(new_job_dict))
                return
        #Remove old jobs
        existed_job_ids = list(self.jobs_by_id.keys())
        for job_id in existed_job_ids:
            if job_id not in current_job_ids:
                del self.jobs_by_id[job_id]
                del self.job_statuses[job_id]
        self.logger.info("Jobs statuses:\n%s" % (str(self.job_statuses)))
        if not active_job_id:
            #self.clouds_job_id = ""
            self.printers_job_id = ""
            self.printing_flag = False
            self.filename = ""
        elif self.printers_job_id != active_job_id:
            self.printers_job_id = active_job_id
            self.printing_flag = True
            active_job_dict = self.jobs_by_id.get(active_job_id)
            if active_job_dict:
                name = active_job_dict.get('name', '')
                if not name:
                    self.logger.warning('HP active job id not in jobs_by_id dictionary')
                else:
                    self.filename = name
            else:
                self.logger.warning('HP active job got no name field')

    def load_and_fill_job_creation_xml(self):
        self.logger.info("Loading xml job template and filling it")
        try:
            with open(self.CREATE_JOB_TEMPLATE_FILE) as f:
                job_template = f.read()
        except (IOError, OSError) as f:
            self.logger.error('Error: cant find template file %s to create job' % self.CREATE_JOB_TEMPLATE_FILE)
        else:
            if not self.materials:
                self.logger.warning('No materials profiles to print')
                return 
            if not self.selected_material_id:
                self.logger.warning('No default material selected. Selection first one.')
                self.selected_material_id = self.default_material_id
            material = self.materials[self.selected_material_id]
            material_name = self.materials[self.selected_material_id].get("name", "")
            #material_xml = self.materials[self.selected_material_id].get("xml", "")
            profiles_dict_for_format = {}
            all_profiles = material.get("profiles", {})
            #self.logger.debug(("Profiles of material: ", pprint.pformat(all_profiles)))
            self.logger.info("...loading xml job template and filling it")
            for profile_type_name in self.JOB_PROFILES_NAMES:
                profiles = all_profiles.get(profile_type_name)
                if not profiles:
                    profile_xml_text = ""
                    self.logger.info("Materials %s got no %s - skipping" % (material.get("name", "Unknown"), profile_type_name))
                else:
                    selected_profile_id = self.selected_profiles_per_material.get(self.selected_material_id, {}).get(profile_type_name, {}).get("id")
                    if not selected_profile_id:
                        for profile in profiles:
                            if profile.get('default', False):
                                self.logger.info("Using default profile %s: %s" % (profile.get('id'), profile.get('Name')))
                                break
                        else:
                            self.logger.warning("No default %s selected, using first last in list" % profile_type_name)
                    else:
                        for profile in profiles:
                            print("%s %s" % (profile_type_name, profile.get('id')))
                            if profile.get('id') == selected_profile_id:
                                self.logger.debug("Using %s %s: %s" % (profile_type_name, profile.get('id'), profile.get('Name')))
                                break
                            else:
                                self.logger.warning('Unable to find %s with selected id %s using first last one' % (profile_type_name, selected_profile_id))
                    profile_xml_text = profile.get("xml", "")
                    #  #TODO remove this code - its for debug only
                    #  with open("/tmp/" + profile_type_name + ".xml", "w") as f:
                    #      f.write(profile_xml_text)
                    if profile_xml_text:
                        self.logger.info("Selecting %s %s" % (profile_type_name[:-1], selected_profile_id))
                    else:
                        self.logger.warning("Unable to select %s with id:%s" % (profile_type_name, selected_profile_id))

                profiles_dict_for_format[profile_type_name] = profile_xml_text
            self.logger.info("...profiles and materials are selected - applying them to a template")
            try:
                job_template = job_template.format(material_id=self.selected_material_id,\
                                   material_name=material_name,\
                                   job_name=self.get_job_name(),\
                                   model_name=self.model_name,\
                                   fw_release_name=self.fw_release_name,\
                                   fw_release_date=self.fw_release_date,\
                                   serial_number=self.serial_number,\
                                   source_host=self.local_ip,\
                                   source_app=self.APP_NAME,\
                                   user_id=self.user_name,\
                                   **profiles_dict_for_format)
            except KeyError as e:
                self.logger.exception("Key error on creation job xml: ")
                self.logger.debug("Job template: %s\nProfiles: %s" % (job_template, profiles_dict_for_format.keys()))
                self.logger.warning("...job xml creation failed")
            else:
                self.logger.info("...job xml creation success")
                self.logger.debug("Job template:\n%s" % job_template)
                #with open("/tmp/my_job_ticket.xml", "w") as f:
                #    f.write(job_template)
                return job_template

    def parse_job_creation_resp(self, resp_text):
        job_id, job_status_url, job_upload_url = "", "", ""
        try:
            resp_text = self.remove_hp_namespace_junk(resp_text)
            job_creation_dict = xmltodict.parse(resp_text)
        except:
            self.logger.error("Error parsing job creation XML")
        else:
            #self.logger.debug("Job creation response:\n%s" % pprint.pformat(job_creation_dict))
            job_creation_dict = job_creation_dict.get("Job", {})
            job_id = job_creation_dict.get("Job_ID", "")
            if not job_id:
                self.logger.warning("Can't get job_id from creation response!")
            job_upload_links = job_creation_dict.get("Links", {}).get("Link", [])
            if not isinstance(job_upload_links, list):
                job_upload_links = [job_upload_links]
            for link_dict in job_upload_links:
                #self.logger.debug("Job creation link: %s" % pprint.pformat(link_dict))
                rel = link_dict.get("@rel")
                if rel == "buildpackage":
                    job_upload_url = link_dict.get('@uri')
                elif rel == "self":
                    job_status_url = link_dict.get('@uri')
        self.logger.info("Got job(%s) status and upload URLs:\n%s\n%s\n" % (job_id, job_status_url, job_upload_url))
        return job_id, job_status_url, job_upload_url

    def create_job(self):
        self.logger.info("Creating a new job...")
        all_services = self.service_discovery(self.device_id, method='Post')
        if all_services:
            create_job_url = all_services.get("jobs")
            if create_job_url:
                debug_job_file_path = getattr(self, 'DEBUG_JOB_XML_PATH', None)
                if debug_job_file_path:
                    with open(debug_job_file_path, "r") as f:
                        job_xml_text = f.read()
                else:
                    job_xml_text = self.load_and_fill_job_creation_xml()
                if not job_xml_text:
                    self.logger.error("Error loading and filling job creation xml")
                    return False
                resp = self.req_post(create_job_url, data=job_xml_text)
                if not resp:
                    return False
                else:
                    status = resp.status_code
                    if status == 200:
                        if resp.text:
                            job_id, _, job_upload_url = self.parse_job_creation_resp(resp.text)
                            #job_id, job_status_url, job_upload_url = self.parse_job_creation_resp(resp.text)
                            if job_id:
                                self.logger.info(f"...job {job_id} created")
                                return job_upload_url
                            else:
                                self.logger.info("...job creation failed")
                                return False
                        else:
                            self.logger.warning('Empty response on job creation')
                            return False
                    elif status == 403:
                        self.parent.register_error(5423, "Job creation failed. Authorization failed", is_blocking=False)
                        return False
                    elif status == 404:
                        self.parent.register_error(5424, "Job creation failed. Job ID %s not found" % job_id, is_blocking=False)
                        return False
                    elif status == 405:
                        self.parent.register_error(5425, "Job creation failed. Automation mode not enable in the printer", is_blocking=False)
                        return False
                    elif status == 409:
                        self.parent.register_error(5429, "Job creation failed. The printer cant print the job", is_blocking=False)
                        return False

    @log.log_exception
    def start_job_when_processing_is_finished(self, buildpackage_status_url, job_id):
        if not buildpackage_status_url:
            self.logger.warning("No build package URL to check status")
            return False
        self.logger.info("Waiting for the build package processing to finish...")
        start_time = time.monotonic()
        while not self.stop_flag:
            bp_dict = {} # otherwise 
            steps = self.UPDATE_LOOP_STEPS
            while steps:
                steps -= 1
                time.sleep(self.UPDATE_LOOP_SLEEP/self.UPDATE_LOOP_STEPS)
                if time.monotonic() - start_time > self.UPLOAD_PROCESSING_TIMEOUT:
                    message = "Timeout waiting for a job %s to be processed"
                    self.logger.warning(message)
                    self.register_error(5999, message, is_blocking=False)
                    return False
                bp_dict = self.req_xml_as_dict(buildpackage_status_url + "/status")
                if not bp_dict:
                    self.logger.debug("Build package status: %s" % pprint.pformat(bp_dict))
                    self.logger.error("Empty response or error on getting build package status")
                    continue
                status_dict = bp_dict.get("BuildPackageStatus", {}).get("UserReported", {}).get("Status", {})
                status = status_dict.get("Message", {}).get("#text", "")
                reason = status_dict.get("Reason", {}).get("#text", "")
                if "Processed" in status:
                    self.logger.info("Build package procession completed. Starting job.")
                    return self.start_job_by_id(job_id)
                elif "Error" in status or "NotProcessed" in status:
                    message = "Printer reported error while processing a build package: %s %s" % (status, reason)
                    self.logger.warning(message)
                    self.logger.debug(bp_dict)
                    self.register_error(5998, message, is_blocking=False)
                    return False
                else:
                    self.logger.debug("Waiting for package processing to finish...")
            self.logger.debug("Build package status: %s" % pprint.pformat(bp_dict))

    def get_buildpackage_status_url(self, text):
        try:
            if not isinstance(text, str):
                self.register_error("Can't use build package id: not string")
            text = self.remove_hp_namespace_junk(text)
            #  if "Job" in text:
            #      pprint.pprint(text)
            build_package_resp_dict = xmltodict.parse(text)
        except:
            self.logger.error("Error parsing build status package XML")
        else:
            links = build_package_resp_dict.get("BuildPackage", {}).get("Links", {}).get("Link", [])
            for link in links:
                self.logger.debug("Got build package link: %s" % pprint.pformat(link))
                if link.get("@rel") == "self":
                    return link.get("@uri")
            else:
                self.logger.warning("Unable to find buildpackage status url in upload response")

    def upload_job_file(self, f, job_upload_url, autostart):
        self.logger.info("Uploading file to the printer...")
        try:
            #import http.client as httpclient
            #httpclient.HTTPConnection.debuglevel = 1
            f.seek(0)
            r = requests.post(job_upload_url, data=f, verify=self.VERIFY_SSL, headers=self.get_auth_headers())
            #httpclient.HTTPConnection.debuglevel = 0
        except requests.RequestException:
            self.logger.warning('Error uploading job file to printer')
            return False
        else:
            if r.status_code == 202:
                self.logger.warning("Job upload success")
                buildpackage_status_url = self.get_buildpackage_status_url(r.text)
                if not buildpackage_status_url:
                    self.logger.warning("Unable got get build package status URL")
                    return False
                if autostart:
                    threading.Thread(target=self.start_job_when_processing_is_finished,\
                            kwargs={"buildpackage_status_url" : buildpackage_status_url}).start()
                return True
            else:
                self.logger.warning("Job upload failed. Status %s. Resp: %s" % (r.status_code, r.text))
                return False

    def start_job_by_id(self, job_id=None):
        if not job_id:
            job_id = self.printers_job_id
        if not job_id:
            self.logger.register_error(5888, "No job provided to start a job")
        self.logger.info("Starting job with ID: %s" % job_id)
        all_services = self.service_discovery(self.device_id, method='Post')
        if all_services:
            jobs_uri = all_services.get("jobs")
            if jobs_uri:
                path = jobs_uri + "%s/start" % job_id
                resp = self.req_post(path)
            if resp:
                status = resp.status_code
                self.logger.warning("Status:%s" % status)
                if status == 200:
                    self.printers_job_id = job_id
                    return True
                elif status == 403:
                    self.parent.register_error(5433, "Job start failed. Authorization failed", is_blocking=False)
                    return False
                elif status == 404:
                    self.parent.register_error(5434, "Job start failed. Job ID %s not found" % job_id, is_blocking=False)
                    return False
                elif status == 405:
                    self.parent.register_error(5435, "Job start failed. Automation mode not enable in the printer", is_blocking=False)
                    return False
                elif status == 409:
                    self.parent.register_error(5439, "Job start failed. The printer cant print the job", is_blocking=False)
                    return False
        self.parent.register_error(5430, "Job start failed. Request to printer failed", is_blocking=False)
        return False

    def set_job_autostart(self, enabled):
        self.logger.info(f"Setting job autostart to {enabled}")
        self.autostart_job = bool(enabled)

    def cancel_job(self, job_id):
        all_services = self.service_discovery(self.device_id, method='Post')
        if all_services:
            jobs_uri = all_services.get("jobs")
            if jobs_uri:
                path = jobs_uri + "/%s/cancel" % job_id
                resp = self.req_post(path)
            status = resp.status_code
            if status == 200:
                return True
            elif status == 403:
                self.parent.register_error(5443, "Cancel failed. Authorization failed", is_blocking=False)
                return False
            elif status == 404:
                self.parent.register_error(5444, "Cancel failed. Job ID %s not found" % job_id, is_blocking=False)
                return False
            elif status == 405:
                self.parent.register_error(5445, "Cancel failed. Automation mode not enable in the printer", is_blocking=False)
                return False
            elif status == 409:
                self.parent.register_error(5449, "Cancel failed. Printer cant cancel the job", is_blocking=False)
                return False
        self.parent.register_error(5430, "Cancel failed. Request to printer failed", is_blocking=False)
        return False

    #TODO possibly is not implemented in the API or disabled by our license
    #TODO recheck this implementation, it is based on pure guess, not docs
    def delete_job(self, job_id=None):
        if not job_id:
            job_id = self.printers_job_id
        if not job_id:
            self.logger.register_error(5888, "No job provided to start a job")
        self.logger.info("Deleting job with ID: %s" % job_id)
        all_services = self.service_discovery(self.device_id, method='Post')
        if all_services:
            jobs_uri = all_services.get("jobs")
            if jobs_uri:
                if not jobs_uri.endswith('/'):
                    jobs_uri += "/"
                jobs_uri = jobs_uri + job_id
                try:
                    resp = requests.request("DELETE", jobs_uri,\
                            headers = {"Authorization" : self.auth_key},\
                            verify=self.VERIFY_SSL, timeout=self.HTTP_TIMEOUT)
                except requests.RequestException as e:
                    self.logger.debug("Error deleting job %s. %s" % (jobs_uri, e))
                else:
                    if resp:
                        status = resp.status_code
                        self.logger.warning("Status:%s" % status)
                        if status == 200 or status == 202:
                            self.printers_job_id = job_id
                            return True
                        elif status == 403:
                            self.parent.register_error(5533, "Job deletion failed. Authorization error", is_blocking=False)
                            return False
                        elif status == 404:
                            self.parent.register_error(5534, "Job deletion failed. Job ID %s not found" % job_id, is_blocking=False)
                            return False
                        elif status == 405:
                            self.parent.register_error(5535, "Job deletion failed. Automation mode not enable in the printer", is_blocking=False)
                            return False
        self.parent.register_error(5430, "Job deletion failed. Request to printer failed", is_blocking=False)
        return False

    def switch_to_material(self, material_id=None, material_name=None):
        if material_id:
            material = self.materials.get(material_id)
            if material:
                self.material = dict(material)
                return True
        elif material_name:
            for material in self.materials:
                if material['name'] == material_name:
                    self.material = dict(material)
                    return True
        else:
            if self.materials:
                for material in self.materials:
                    self.material = dict(material)
                    break
            if self.material:
                self.logger.error("No material ID or material name provided. Fallbacking to %s" % self.material["name"])
            else:
                self.logger.error("No material ID or material name provided. Cant print")
        return False

    def get_profiles_from_material_xml(self, xml_text):
        found_profiles = {}
        current_profile = ""
        line_number = 0
        for line in xml_text.split("\n"):
            line_number += 1
            if not current_profile:
                for profile_name in self.JOB_PROFILES_NAMES:
                    profile_name = profile_name.strip("s")
                    profile_name_tag = ":" + profile_name + ">"
                    if profile_name_tag in line:
                        current_profile = profile_name 
                        if not profile_name + "s" in found_profiles:
                            found_profiles[profile_name + "s"] = ["",]
                        else:
                            found_profiles[profile_name + "s"].append("")
                        #self.logger.debug("Found start of %s. L:%d" % (profile_name, line_number))
            if current_profile:
                found_profiles[current_profile + "s"][-1] += line + "\n"
                profile_name_tag = ":" + current_profile + ">"
                if profile_name_tag in line and "</" in line:
                    #self.logger.debug('Found end of %s. L:%d ' % (current_profile, line_number))
                    current_profile = ""
        return found_profiles

    def request_sensors(self):
        sensors_dict = self.get_service_as_dict('sensors')
        if sensors_dict:
            self.sensors_data = sensors_dict
            #self.logger.debug("Sensors: %s" % pprint.pformat(sensors_dict))
        else:
            self.logger.warning("Error updating sensors")

    #NOTE untested and completely raw
    def request_telemetry(self):
        #TODO finish this code when we will get a license for telemetry - now it is unfinished
        self.logger.debug("Requesting telemetry...")
        telemetry_url = f"{self.deviceproxy_host}/devices/{self.device_id}/telemetry"
        resp = self.req(telemetry_url, post=True) #redirects are enabled by default
        if resp:
            self.logger.debug(f"Telemetry {resp.status} {resp.body}")
            return resp.body
        self.logger.debug("done")

    #NOTE untested and completely raw
    def request_telemetry_for_current_job(self, job_id=None):
        self.logger.debug("Requesting telemetry...")
        if not job_id:
            job_id = self.printers_job_id
        if not job_id:
            self.logger.warning("No printer's job id to request telemetry")
            return
        telemetry_url = f"{self.deviceproxy_host}/devices/{self.device_id}/jobs/{job_id}/telemetry"
        resp = self.req(telemetry_url, post=True) #redirects are enabled by default
        if resp:
            self.logger.debug(f"Telemetry {resp.status} {resp.body}")
            return resp.body
        self.logger.debug("done")

    def close(self):
        self.logger.info("Closing " + __name__)
        self.stop_flag = True


